@@ -81,7 +81,7 @@ import fnmatch
 import json
 
 if len(sys.argv) == 1:
-  print('''Usage: file_packager.py TARGET [--preload A...] [--embed B...] [--exclude C...] [--no-closure]] [--js-output=OUTPUT.js] [--no-force] [--use-preload-cache] [--no-heap-copy] [--separate-metadata]
+  print('''Usage: file_packager.py TARGET [--preload A [B..]] [--embed C [D..]] [--exclude E [F..]]] [--js-output=OUTPUT.js] [--no-force] [--use-preload-cache] [--indexedDB-name=EM_PRELOAD_CACHE] [--no-heap-copy] [--separate-metadata] [--lz4] [--use-preload-plugins]
 See the source for more details.''')
   sys.exit(0)
 
@@ -115,11 +115,11 @@ force = True
 use_preload_cache = False
 indexeddb_name = 'EM_PRELOAD_CACHE'
 # If set to True, the blob received from XHR is moved to the Emscripten HEAP,
-# optimizing for mmap() performance.
+# optimizing for mmap() performance (if ALLOW_MEMORY_GROWTH=0).
 # If set to False, the XHR blob is kept intact, and fread()s etc. are performed
 # directly to that data. This optimizes for minimal memory usage and fread()
 # performance.
-no_heap_copy = True
+heap_copy = True
 # If set to True, the package metadata is stored separately from js-output
 # file which makes js-output file immutable to the package content changes.
 # If set to False, the package metadata is stored inside the js-output file
@@ -146,7 +146,7 @@ for arg in sys.argv[2:]:
     indexeddb_name = arg.split('=', 1)[1] if '=' in arg else None
     leading = ''
   elif arg == '--no-heap-copy':
-    no_heap_copy = False
+    heap_copy = False
     leading = ''
   elif arg == '--separate-metadata':
     separate_metadata = True
@@ -496,7 +496,7 @@ for file_ in data_files:
 if has_preloaded:
   if not lz4:
     # Get the big archive and split it up
-    if no_heap_copy:
+    if heap_copy:
       use_data = '''
         // copy the entire loaded file into a spot in the heap. Files will refer to slices in that. They cannot be freed though
         // (we may be allocating before malloc is ready, during startup).
@@ -529,7 +529,7 @@ if has_preloaded:
     os.unlink(temp)
     use_data = '''
           var compressedData = %s;
-          compressedData.data = byteArray;
+          compressedData['data'] = byteArray;
           assert(typeof LZ4 === 'object', 'LZ4 not present - was your app build with  -s LZ4=1  ?');
           LZ4.loadPackage({ 'metadata': metadata, 'compressedData': compressedData });
           Module['removeRunDependency']('datafile_%s');
@@ -806,10 +806,8 @@ if has_preloaded:
             function(useCached, metadata) {
               Module.preloadResults[PACKAGE_NAME] = {fromCache: useCached};
               if (useCached) {
-                console.info('loading ' + PACKAGE_NAME + ' from cache');
                 fetchCachedPackage(db, PACKAGE_PATH + PACKAGE_NAME, metadata, processPackageData, preloadFallback);
               } else {
-                console.info('loading ' + PACKAGE_NAME + ' from remote');
                 fetchRemotePackage(REMOTE_PACKAGE_NAME, REMOTE_PACKAGE_SIZE,
                   function(packageData) {
                     cacheRemotePackage(db, PACKAGE_PATH + PACKAGE_NAME, packageData, {uuid:PACKAGE_UUID}, processPackageData,
